/**
  *  \file client/cm-standard.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>                // sin, cos
#include "client/cm-standard.h"
#include "game/formula.h"       // PI
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "client/chartloc.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "ui/ui.h"
#include "client/screens.h"
#include "game/ship.h"
#include "gfx/update.h"
#include "int/intglobal.h"
#include "int/if/shipif.h"
#include "int/if/planetif.h"
#include "int/keymap.h"
#include "game/planet.h"
#include "int/simple.h"

/************************** START Prototype Code *************************/

static void drawFancyBar(GfxCanvas& can, GfxRect r);

class WStarchartTileSkin : public UISkin {
 public:
    const UIColorSet& getColors() const;
    void  outText(GfxCanvas& can,  int f, int x, int y, string_t text);
    void  outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text);
    void  drawBackground(GfxCanvas& can, GfxRect r);
};

const UIColorSet&
WStarchartTileSkin::getColors() const
{
    return DARK_COLOR_SET;
}

void
WStarchartTileSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = GFX_TRANSPARENT;
    ::outText(can, parm, x, y, text);
}

void
WStarchartTileSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = GFX_TRANSPARENT;
    ::outTextF(can, parm, x, y, lim, text);
}

void
WStarchartTileSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    // empty
}

class WStarchartTileContainer : public UIBaseWidget {
    ObjectHolder h;
    WStarchartTileSkin skin;
 public:
    WStarchartTileContainer();

    virtual void          drawContent(GfxCanvas& canvas);                       // draw content of widget, end-users override this
    virtual void          drawEverything(GfxCanvas& canvas);                    // draw everything including children, bases override this
    virtual void          drawBelow(UIBaseWidget& child, GfxCanvas& canvas);    // draw everything below child, bases override this
    virtual GfxCanvas&    getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr); // get canvas to draw child on, bases override this
    virtual bool          handleEvent(const UIEvent& event, bool second_pass);   // can be overridden by whoever needs it, base provides default
    virtual void          onAddRemoveChild(UIBaseWidget& the_widget, bool added); // overridden in base
    virtual void          getLayoutInfo(LayoutInfo&);                          // overridden by everyone
    virtual void          onResize(GfxRect oldsize, bool by_layout);           // overridden in base
    virtual void          onChildResize(UIBaseWidget& child, GfxRect oldsize); // overridden in base
    virtual int           run();                 // overridden in root
    virtual void          stop(int arg);         // overridden in root

    void doLayout(GChartViewport& viewport);
    void addNewTile(UIBaseWidget* w);
};

WStarchartTileContainer::WStarchartTileContainer()
{
    setSkin(skin);
}

void
WStarchartTileContainer::drawContent(GfxCanvas& canvas)
{
    drawFancyBar(canvas, getExtent());
    canvas.update(getExtent());
}

void
WStarchartTileContainer::drawEverything(GfxCanvas& canvas)
{
    drawContent(canvas);
    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
        p->drawEverything(canvas);
}

void
WStarchartTileContainer::drawBelow(UIBaseWidget& child, GfxCanvas& canvas)
{
    // draw everything below child, bases override this
}

GfxCanvas&
WStarchartTileContainer::getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr)
{
    // get canvas to draw child on, bases override this
}

bool
WStarchartTileContainer::handleEvent(const UIEvent& event, bool second_pass)
{
    // can be overridden by whoever needs it, base provides default
    return false;
}

void
WStarchartTileContainer::onAddRemoveChild(UIBaseWidget& the_widget, bool added)
{
    /* FIXME: at first, we assume no uncontrolled add/remove.
       Otherwise, we'd simply have to request relayout.  */
}

void
WStarchartTileContainer::getLayoutInfo(LayoutInfo& li)
{
    li.min_size = li.pref_size = GfxPoint(0, 0);
    li.flags = li_NoLayout;
}

void
WStarchartTileContainer::onResize(GfxRect oldsize, bool by_layout)
{
    // Nothing to do?
}

void
WStarchartTileContainer::onChildResize(UIBaseWidget& child, GfxRect oldsize)
{
    /* FIXME: at first, we assume no uncontrolled add/remove/resize.
       Otherwise, we'd simply have to request relayout.  */
}

int
WStarchartTileContainer::run()
{
    /* does not run itself */
    return 0;
}

void
WStarchartTileContainer::stop(int /*arg*/)
{ /* does not run itself */ }

void
WStarchartTileContainer::doLayout(GChartViewport& viewport)
{
    enum { EDGE = 5 };

    // Compute total height and width
    int total_height = EDGE;
    int min_width    = 0;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        total_height += EDGE + li.pref_size.y;
        if (li.pref_size.x > min_width)
            min_width = li.pref_size.x;
    }
    min_width += 2*EDGE;

    // Set own size and allocate space
    GfxRect myExtent(0, 0, min_width, total_height);
    myExtent.moveToEdge(viewport.view, 2, 1, 10);
    setExtent(myExtent, true);

    // Move children
    int y = myExtent.y + EDGE;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        w->setExtent(GfxRect(myExtent.x + EDGE, y, myExtent.w - 2*EDGE, li.pref_size.y), false);
        y += li.pref_size.y + EDGE;
    }
}

void
WStarchartTileContainer::addNewTile(UIBaseWidget* w)
{
    addChildWidget(h.add(w), getLastChild());
}

/*************************** END Prototype Code **************************/

static const int OBJS_PER_PAGE = 15;

static void
drawFancyBar(GfxCanvas& can, GfxRect r)
{
    const int icolor = standard_colors[COLOR_SHIELD+3];
    const int fcolor = standard_colors[COLOR_SHIELD+5];
    const int z = standard_colors[COLOR_BLACK];

    int alpha;

    if (ui_screen_bits <= 8) {
        can.drawBar(r, icolor, z, GRAY50_FILL, GFX_ALPHA_OPAQUE);
        alpha = GFX_ALPHA_OPAQUE;
    } else {
        can.drawBar(r, icolor, z, SOLID_FILL, 128);
        alpha = (3*GFX_ALPHA_OPAQUE + GFX_ALPHA_TRANSPARENT)/4;
    }

    can.drawBar(GfxRect(r.x-1,   r.y-2,   r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-1,   r.y+r.h, r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-2,   r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x+r.w, r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y+r.h), fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y+r.h), fcolor, alpha);
}

static void
drawChartButton(GfxCanvas& can, int& x, int y, const string_t& text)
{
    int width = 8 + fonts[FONT_SMALL]->getTextWidth(text);
    UIButton::drawButton(can, GfxRect(x, y, width, 15), text, UIButton::bf_SmallSize);
    x += width + 1;
}

static void
drawCursor(GfxCanvas& can, GfxPoint pt, int angle)
{
    static const int SIZE = 15;
    static const double FACTOR = 5 * PI / 180;

    angle = -angle;

    GfxParameters parm;
    parm.color = standard_colors[COLOR_RED];

    /* it turns out that scaling the cursor doesn't look that good. */
    const double scaled_size = SIZE;

    for (int i = 0; i < 2; ++i) {
        double d = angle * PI / 180;
        drawLine(can, parm,
                 GfxPoint(pt.x + roundToInt(4*std::sin(d)), pt.y + roundToInt(4*std::cos(d))),
                 GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d)), pt.y + roundToInt(scaled_size*std::cos(d))));
        for (int j = -8; j < 8; ++j) {
            drawLine(can, parm,
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j))),
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j+FACTOR)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j+FACTOR))));
        }
        angle += 180;
    }
}

static void
drawObjectList(GfxCanvas& can, GChartLocation& loc)
{
    const GObjectList& chart_objlist = loc.getObjectList();

    if (!chart_objlist.getNumObjects())
        return;

    int lines = chart_objlist.getNumObjects();
    if (lines > OBJS_PER_PAGE)
        lines = OBJS_PER_PAGE;

    GfxPoint pt = loc.getViewport().getCenter();
    int h = font_heights[FONT_NORMAL]*lines;
    int y = pt.y - h/2;

    drawFancyBar(can, GfxRect(pt.x - 300, y-2, 280, h+4));

    GfxParameters parm;
    for (int i = 0; i < lines; ++i) {
        GObjectId id = chart_objlist.getObjectId(i + 1);
        UIColor::Color color = (id.isGameObject()
                                ? getPlayerColor(id.getObject().getOwner())
                                : UIColor::tc_Static);
        if (id == loc.getCurrentObject()) {
            if (ui_screen_bits <= 8)
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+4], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
            else
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+8], 0, SOLID_FILL, 128);
            parm.color = standard_colors[BLACK_COLOR_SET[color]];
            parm.bgcolor = GFX_TRANSPARENT;
            parm.alpha = GFX_ALPHA_OPAQUE;
        } else {
            parm.color = standard_colors[BLACK_COLOR_SET[color]];
            parm.bgcolor = GFX_TRANSPARENT;
            parm.alpha = GFX_ALPHA_OPAQUE;
        }
        parm.font = fonts[FONT_NORMAL];
        outTextF(can, parm, pt.x-298, y+i*font_heights[FONT_NORMAL], 276,
                 id.getName());
    }
}

static void
drawObject(GfxCanvas& can, GChartLocation& chart_location)
{
    (void) can;
    (void) chart_location;
//     if (!chart_objlist.getNumObjects())
//         return;

//     GObjectId id = chart_objlist.getObjectId(chart_location.ol_pos);
//     if (GShip* s = dynamic_cast<GShip*>(&id.getObject())) {
//         GfxPoint pt = chart_viewport.getCenter();
//         GfxParameters parm;
//         int h = 110;
//         int y = pt.y - h/2;
//         int hull = s->getHullId();
//         int x;

//         if (hull > 0) {
//             drawFancyBar(can, GfxRect(pt.x + 20, y, 280, h));

//             GfxRect imgrect(pt.x + 25, y + 5, 100, 100);
//             Ptr<GfxPixmap> pix = requestPixmap(ResId(res::ship, getHull(hull).getPicture(), hull));
//             if (pix)
//                 pix->blitSized(can, imgrect.x, imgrect.y, imgrect.w, imgrect.h);

//             imgrect.grow(1, 1);
//             parm.color = standard_colors[COLOR_SHIELD+5];
//             parm.alpha = 192;
//             drawRectangle(can, parm, imgrect);

//             x = pt.x + 130;

//             drawChartButton(can, x, y + 90, "F1");
//             drawChartButton(can, x, y + 90, "F6");
//             drawChartButton(can, x, y + 90, "Ins");

//             x = pt.x + 130;
//         } else {
//             drawFancyBar(can, GfxRect(pt.x + 20, y, 180, h));
//             x = pt.x + 30;
//         }
//         parm.color = standard_colors[COLOR_GRAY];
//         parm.alpha = GFX_ALPHA_OPAQUE;
//         parm.bgcolor = GFX_TRANSPARENT;
//         parm.font = fonts[FONT_NORMAL];
//         outTextF(can, parm, x, y + 5, 150, s->getName());
//         if (hull > 0)
//             outTextF(can, parm, x, y + 5 + font_heights[FONT_NORMAL], 150,
//                      format(_("a %s %s"), player_racenames.getAdjName(s->getOwner()), s->getHull().getShortName()));
//         else
//             outTextF(can, parm, x, y + 5 + font_heights[FONT_NORMAL], 150,
//                      format(_("a %s starship"), player_racenames.getAdjName(s->getOwner())));
//     }
}

/*************************** WStandardChartMode **************************/

/** Constructor. */
WStandardChartMode::WStandardChartMode()
    : WChartMode(ChartMode_BaseMode),
      cursor_angle(0),
      overlays(0)
{ }

/** Destructor. */
WStandardChartMode::~WStandardChartMode()
{
    delete overlays;
}

void
WStandardChartMode::onMove(GChartLocation& loc)
{
    // Build overlay tile
    // delete overlays;
    // overlays = 0;
    // if (loc.getCurrentObject().isValid()) {
    //     if (GShip* sh = dynamic_cast<GShip*>(&loc.getCurrentObject().getObject())) {
    //         global_anyship_type.setCurrent(sh->getId());
    //         overlays = new WStarchartTileContainer();
    //         overlays->addNewTile(createShipScreenHeaderTile(global_anyship_type));
    //         overlays->doLayout(loc.getViewport());
    //     }
    // }

    // Call next mode
    if (WChartMode* cm = getNextMode())
        cm->onMove(loc);
}

void
WStandardChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation().getNearestTo(loc.getCenter()));
    if (!loc.getCurrentObject().isValid()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_BLUE];
        drawHLine(can, parm, sc.x-30, sc.y, sc.x-6);
        drawHLine(can, parm, sc.x+30, sc.y, sc.x+6);
        drawVLine(can, parm, sc.x, sc.y-30, sc.y-6);
        drawVLine(can, parm, sc.x, sc.y+30, sc.y+6);
    } else {
        ::drawCursor(can, sc, cursor_angle);
    }
}

void
WStandardChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    drawObjectList(can, loc);
    // drawObject(can, loc);

    if (overlays) {
        GfxUpdateOptimizer o(can);
        GfxRect r;
        overlays->drawEverything(o);
        o.clearUpdateRegion(r); /* we forget the region, because the whole screen is refreshed afterwards */
    }

    // Coordinates
    GPoint pt = loc.getLocation().getNormalized();
    GfxParameters parm;
    GfxRect view(loc.getViewport().view);
    parm.color = standard_colors[COLOR_WHITE];
    parm.just_x = parm.just_y = 2;
    parm.font = fonts[FONT_FIXED];
    outText(can, parm, view.getRightX(), view.getBottomY(), format("%4d,%4d", pt.x, pt.y));

    // Sector Number
    int secn = pt.getSectorNumber();
    if (secn != 0) {
        parm.just_x = 0;
        outText(can, parm, view.x, view.getBottomY(), format(_("Sector %d"), secn));
    }
}

bool
WStandardChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Tick) {
        if (loc.getCurrentObject().isValid()) {
            /* Advance rotation, and trigger redraw. This used to draw
               in-place; the new way supports real interaction with
               other overlays. */
            cursor_angle = (cursor_angle + 3) % 360;
            GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation());
            getParent()->markDirty(GfxRect(sc.x - 15, sc.y - 15, 31, 31));
        }
    } else if (event.type == ev_Key) {
        // Check keymap
        IntKeymapRef km = IntKeymap::getKeymapByName("STARCHART");
        if (loc.getCurrentObject().isValid()) {
            if (dynamic_cast<GShip*>(&loc.getCurrentObject().getObject())) {
                km = IntKeymap::getKeymapByName("SHIPLOCK");
            } else if (dynamic_cast<GPlanet*>(&loc.getCurrentObject().getObject())) {
                km = IntKeymap::getKeymapByName("PLANETLOCK");
            } else {
                // No object
            }
        }

        // Do we have a key?
        uint32_t cmd;
        if (km && isAtom(cmd = km->lookupCommand(event.key.code))) {
            IntTemporaryContextLock lock;
            if (loc.getCurrentObject().isValid()) {
                if (dynamic_cast<GShip*>(&loc.getCurrentObject().getObject())) {
                    global_contexts.push_back(new IntShipContext(loc.getCurrentObject().getId()));
                } else if (dynamic_cast<GPlanet*>(&loc.getCurrentObject().getObject())) {
                    global_contexts.push_back(new IntPlanetContext(loc.getCurrentObject().getId()));
                } else {
                    // No object
                }
            }
            executeStatement(atomStr(cmd), 0);
            return true;
        }
    }
    return false;
}
