/**
  *  \file game/cspec.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/cspec.h"
#include "cpluslib/format.h"

/* We require particular values for the el_foo constants. */
CTASSERT(el_Neutronium == 0);
CTASSERT(el_Tritanium == 1);
CTASSERT(el_Duranium == 2);
CTASSERT(el_Molybdenum == 3);
CTASSERT(el_Fighters == 4);
CTASSERT(el_Colonists == 5);
CTASSERT(el_Supplies == 6);
CTASSERT(el_Money == 7);
CTASSERT(el_Torps == 8);

static const char cargo_type_letters[]    = "NTDMFCS$W";
static const char cargo_type_letters_lc[] = "ntdmfcs$w";


/** Construct blank cargospec. */
GCargoSpec::GCargoSpec()
{
    clear();
}

/** Construct from GCost. */
GCargoSpec::GCargoSpec(const GCost& c)
{
    clear();
    amounts[el_Tritanium]  = c.get(el_Tritanium);
    amounts[el_Duranium]   = c.get(el_Duranium);
    amounts[el_Molybdenum] = c.get(el_Molybdenum);
    amounts[el_Supplies]   = c.get(el_Supplies);
    amounts[el_Money]      = c.get(el_Money);
}

/** Parse number from string.
    \param value [in] the string
    \param i [in/out] current parse position
    \return the number */
static int32
eatNumber(const string_t& value, string_t::size_type& i)
{
    int32 amount = 0;
    while (i < value.size() && value[i] >= '0' && value[i] <= '9') {
        amount = 10*amount + value[i] - '0';
        ++i;
    }
    return amount;
}

/** Construct from CargoSpec or PHost string. */
GCargoSpec::GCargoSpec(const string_t& str)
{
    clear();

    /* Like PCC 1.x, we accept cargospecs and PHost format:
          Cargospec:  123TDM
          PHost:      T123 D123 M123 */
    string_t::size_type i = 0;
    while (1) {
        /* skip whitespace */
        while (i < str.size() && (str[i] == ' ' || str[i] == '\t'))
            ++i;
        if (i >= str.size())
            break;

        if (str[i] >= '0' && str[i] <= '9') {
            /* Cargospec */
            int32_t amount = eatNumber(str, i);
            int type;
            while (i < str.size() && (type = charToCargoType(str[i])) >= 0) {
                add(GCargoType(type), amount);
                ++i;
            }
        } else {
            /* Must be PHost format */
            int type = charToCargoType(str[i]);
            ++i;
            if (type < 0 || i >= str.size()) {
                /* at this point, we know that it has a wrong format.
                   FIXME: what can we do? Should we report an error?
                   PCC 1.x just exits, so we now do that as well. */
                break;
            }
            add(GCargoType(type), eatNumber(str, i));
        }
    }
}

/** Add cargospecs. */
GCargoSpec&
GCargoSpec::operator+=(const GCargoSpec& other)
{
    for (int i = 0; i < MAX; ++i)
        amounts[i] += other.amounts[i];
    return *this;
}

/** Subtract cargospecs. */
GCargoSpec&
GCargoSpec::operator-=(const GCargoSpec& other)
{
    for (int i = 0; i < MAX; ++i)
        amounts[i] -= other.amounts[i];
    return *this;
}

/** Multiply cargospecs. */
GCargoSpec&
GCargoSpec::operator*=(int32_t count)
{
    for (int i = 0; i < MAX; ++i)
        amounts[i] *= count;
    return *this;
}

/** Convert to PHost string. Note that this returns a valid string
    accepted by PHost for a "cost" option only if its content is
    restricted to positive amounts of T/D/M/S/$; "beamup" accepts
    everything but W (el_Torps). */
string_t
GCargoSpec::toPHostString() const
{
    string_t result;
    for (int i = 0; i < MAX; ++i) {
        if (amounts[i] != 0) {
            if (result.size())
                result += ' ';
            result += cargo_type_letters[i];
            result += format("%d", amounts[i]);
        }
    }
    if (result.empty()) {
        result = "S0";
    }
    return result;
}

/** Convert to CargoSpec string. */
string_t
GCargoSpec::toCargoSpecString() const
{
    string_t result;
    GCargoTypeSet did;

    /* Special case: if it has T=D=M, combine these */
    if (amounts[el_Tritanium] != 0 && amounts[el_Tritanium] == amounts[el_Duranium] && amounts[el_Tritanium] == amounts[el_Molybdenum]) {
        result += format("%dTDM", amounts[el_Tritanium]);
        did += el_Tritanium;
        did += el_Molybdenum;
        did += el_Duranium;
    }

    /* Regular processing */
    for (int i = 0; i < MAX; ++i) {
        if (!did.contains(GCargoType(i)) && amounts[i] != 0) {
            if (!result.empty())
                result += ' ';
            result += format("%d%c", amounts[i], cargo_type_letters[i]);
        }
    }
    return result;
}

/** Convert to cost. This discards everything not accepted by GCost. */
GCost
GCargoSpec::toCost() const
{
    GCost rv;
    rv.set(el_Tritanium,  amounts[el_Tritanium]);
    rv.set(el_Duranium,   amounts[el_Duranium]);
    rv.set(el_Molybdenum, amounts[el_Molybdenum]);
    rv.set(el_Supplies,   amounts[el_Supplies]);
    rv.set(el_Money,      amounts[el_Money]);
    return rv;
}

/** Set component value.
    \param t component
    \param amount amount to use */
void
GCargoSpec::set(GCargoType t, int32_t amount)
{
    ASSERT(isValidType(t));
    amounts[t] = amount;
}

/** Get component value.
    \param t component */
int32_t
GCargoSpec::get(GCargoType t) const
{
    ASSERT(isValidType(t));
    return amounts[t];
}

/** Add component.
    \param t component
    \param amount amount to add */
void
GCargoSpec::add(GCargoType t, int32_t amount)
{
    ASSERT(isValidType(t));
    amounts[t] += amount;
}

/** Remove component.
    \param t component
    \param amount amount to remove */
void
GCargoSpec::remove(GCargoType t, int32_t amount)
{
    ASSERT(isValidType(t));
    amounts[t] -= amount;
}

/** Clear. Sets all components to zero. */
void
GCargoSpec::clear()
{
    for (int i = 0; i < MAX; ++i)
        amounts[i] = 0;
}

/** Check whether this cargospec contains at least as much as required for \c other.
    Supply sale is taken into account. */
bool
GCargoSpec::isEnoughFor(const GCargoSpec& other) const
{
    // Subtract, and perform supply sale
    GCargoSpec tmp(*this);
    tmp -= other;
    tmp.doSupplySale();

    // Still positive?
    return tmp.isValid();
}

/** Check whether this is a valid cargo amount, i.e.\ all values non-negative. */
bool
GCargoSpec::isValid() const
{
    for (int i = 0; i < MAX; ++i)
        if (amounts[i] < 0)
            return false;
    return true;
}

/** Perform supply sale. If this contains a negative amount of money,
    try to make it positive by selling supplies. */
void
GCargoSpec::doSupplySale()
{
    if (amounts[el_Money] < 0) {
        if (amounts[el_Supplies] >= -amounts[el_Money]) {
            // Enough supplies to cover this money shortage
            amounts[el_Supplies] += amounts[el_Money];
            amounts[el_Money] = 0;
        } else {
            // Not enough supplies, so sell as much as possible
            amounts[el_Money] += amounts[el_Supplies];
            amounts[el_Supplies] = 0;
        }
    }
}

/** In-place divide by integer. Splits this cargospec evenly into n
    parts, such that (resultValue*=n).isEnoughFor(originalValue).
    \retval true on success
    \retval false if this divides by zero */
bool
GCargoSpec::divide(int32_t n)
{
    if (n == 0)
        return 0;

    for (int i = 0; i < MAX; ++i)
        amounts[i] /= n;
    return true;
}

/** Divide by cargospec. Computes how many times \c cs fits into \c *this,
    such that (cs*=result).isEnoughFor(*this). Alternatively, if this is
    the amount of cargo available, and \c cs is a part cost, computes the
    number of parts that can be built. Supply sale is taken into account.
    \param cs [in] Part cost
    \param result [out] Part count
    \retval true on success
    \retval false if this divides by zero (cs is empty) */
bool
GCargoSpec::divide(const GCargoSpec& cs, int32_t& result) const
{
    bool did = false;

    // Handle most
    for (int i = 0; i < MAX; ++i) {
        if (i != el_Money && cs.amounts[i] != 0) {
            int32_t q = amounts[i] / cs.amounts[i];
            if (!did || q < result) {
                did = true;
                result = q;
            }
        }
    }

    // Handle money and supply sale
    int32_t moneyCost = cs.amounts[el_Money] + cs.amounts[el_Supplies];
    if (moneyCost != 0) {
        int32_t q = (amounts[el_Money] + amounts[el_Supplies]) / moneyCost;
        if (!did || q < result) {
            did = true;
            result = q;
        }
    }

    return did;
}

/** Check for valid type.
    \return true iff t is a valid index for get/set/add/... */
bool
GCargoSpec::isValidType(GCargoType t)
{
    return (t >= 0 && t <= el_Torps);
}

/** Convert character into cargo type.
    \return el_XXX on success, -1 on failure. */
int
GCargoSpec::charToCargoType(char c)
{
    using std::strchr;
    if (const char* p = strchr(cargo_type_letters, c))
        return p - cargo_type_letters;
    if (const char* p = strchr(cargo_type_letters_lc, c))
        return p - cargo_type_letters_lc;
    return -1;
}
