/**
  *  \file resmgr/resmgr.h
  *  \brief Resource Manager
  *
  *  This is the interface of the resource manager, provider and
  *  cache.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_RESMGR_RESMGR_H
#define PCC_V2_RESMGR_RESMGR_H

#include <map>
#include <list>
#include "cpluslib/misc.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "cpluslib/holder.h"
#include "gfx/anim.h"
#include "gfx/pixmap.h"
#include "io/dirs.h"
#include "io/streams.h"
#include "sound/sound.h"
#include "cpluslib/ptrvector.h"

class ResManager;

/** Resource identifier. A resource is identified by a name and up to
    two numbers. For example, ResId("vcr.ftr", 2), stringified as
    "vcr.ftr2", is the resource Id for player 2's fighter in the VCR.
    A resource can be generalized by removing numbers. If there were
    no picture for "vcr.ftr2", we'd try "vcr.ftr". */
class ResId {
    const char* name;
    int         id1;
    int         id2;
 public:
    /** Construct null resource Id. */
    ResId()
        : name(0), id1(0), id2(0) { }
    /** Construct resource Id.
        \param name    resource name, must be statically allocated string
        \param id1,id2 identifiers */
    ResId(const char* name, int id1 = 0, int id2 = 0)
        : name(name), id1(id1), id2(id2) { }
    /** True iff this is a generic resource Id. A generic
        resource Id cannot be further generalized. */
    bool isGeneric() const
        { return id1 == 0 && id2 == 0; }
    void generalize();

    /** Get resource name. */
    const char* getName() const
        { return name; }
    /** Get first Id. 0 if not set. */
    int getId1() const
        { return id1; }
    /** Get second Id. 0 if not set. */
    int getId2() const
        { return id2; }
    string_t toString() const;

    bool operator==(const ResId& rhs) const;
    bool operator<(const ResId& rhs) const;
    bool operator>(const ResId& rhs) const
        { return rhs < *this; }
    bool operator<=(const ResId& rhs) const
        { return !(rhs < *this); }
    bool operator>=(const ResId& rhs) const
        { return !(*this < rhs); }
    bool operator!=(const ResId& rhs) const
        { return !(*this == rhs); }

    bool isElem(const char* name, int level) const;
};

class ResProvider : public Object {
    friend class ResManager;
    ResManager* manager;
    void detach();
 public:
    /** Create empty resource provider. */
    ResProvider()
        : manager(0)
        { }
    ~ResProvider()
        { detach(); }
    /** Load requested pixmap. Called by resource manager if user
        requested the specified pixmap. This function should load the
        pixmap if it can and call mgr.addPixmap() if possible; it does
        nothing if it cannot satisfy the request. If multiple pixmaps
        are available, add them all. */
    virtual void loadPixmap(ResManager& mgr, ResId id) = 0;
    /** Load requested sound. Called by resource manager if user
        requested the specified sound. This function should load the
        sound if it can and call mgr.addSound() if possible; it does
        nothing if it cannot satisfy the request. If multiple sounds
        are available, add them all. */
    virtual void loadSound(ResManager& mgr, ResId id) = 0;
    /** Load requested animation. Called by resource manager if user
        requested the specified animation. This function should load
        the animation if it can and call mgr.addAnimation() if
        possible; it does nothing if it cannot satisfy the request. If
        multiple animations are available, add them all. */
    virtual void loadAnimation(ResManager& mgr, ResId id) = 0;
    /** Load requested font. Called by resource manager if user
        requested the specified font. This function should return the
        font if it can provide it, 0 otherwise. */
    virtual Ptr<GfxFont> getFont(ResId id) = 0;
    /** Open a stream. Called by resource manager if user requested a
        stream. This function should return the stream open for
        reading in binary mode if it can, 0 otherwise. There is no
        required relation between streams and other resources, it need
        not be possible to request the underlying stream for a pixmap,
        for example. */
    virtual Ptr<Stream> getStream(ResId id) = 0;
};

//! Simple resource provider.
class ResSimpleProvider : public ResProvider {
 public:
    ResSimpleProvider()
        { }
    void loadPixmap(ResManager& mgr, ResId id);
    void loadSound(ResManager& mgr, ResId id);
    void loadAnimation(ResManager& mgr, ResId id);
    Ptr<GfxFont> getFont(ResId id);
    Ptr<Stream> getStream(ResId id);
};

//! Resources from a directory (dir:NAME).
class ResDirectory : public ResProvider {
    /// Alias table entry.
    struct Node {
        string_t name;
    };
    /// Type for alias definitions.
    typedef std::map<string_t, Node> alias_map_t;
    /// Alias definitions.
    static alias_map_t alias_map;

    /// Underlying directory.
    Ptr<Directory> dir;

    void loadAliases(Stream& s);
    Ptr<Stream> openStream(string_t name, const char*const* suffixes);
 public:
    ResDirectory(Ptr<Directory> dir);
    ~ResDirectory();

    void loadPixmap(ResManager& mgr, ResId id);
    void loadSound(ResManager& mgr, ResId id);
    void loadAnimation(ResManager& mgr, ResId id);
    Ptr<GfxFont> getFont(ResId id);
    Ptr<Stream> getStream(ResId id);
};

//! Resource manager.
class ResManager {
    /// Pixmap cache entry.
    struct PixmapNode {
        int use_count;
        int pixmap_count;
        std::list<Ptr<GfxPixmap> > pixmaps;
        PixmapNode()
            : use_count(1), pixmap_count(0) { }
    };
    /// Sound cache entry.
    struct SoundNode {
        int use_count;
        int sound_count;
        ptr_vector<SfxSampleData> sounds;
        SoundNode()
            : use_count(1), sound_count(0) { }
        ~SoundNode();
    };
    /// Animation cache entry.
    struct AnimationNode {
        int use_count;
        int anim_count;
        std::list<Ptr<GfxAnimation> > animations;
        AnimationNode()
            : use_count(1), anim_count(0) { }
    };

    /// Pixmap cache.
    typedef std::map<ResId, PixmapNode*> pixmap_cache_t;
    pixmap_cache_t pixmap_cache;

    /// Sound cache.
    typedef std::map<ResId, SoundNode*> sound_cache_t;
    sound_cache_t sound_cache;

    /// Animation cache.
    typedef std::map<ResId, AnimationNode*> anim_cache_t;
    anim_cache_t anim_cache;

    /// Resource provider list.
    typedef std::list<ResProvider*> provider_list_t;
    provider_list_t provider_list;

    static ResManager instance;

    PixmapNode*    getPixmapNode(ResId id);
    SoundNode*     getSoundNode(ResId id);
    AnimationNode* getAnimationNode(ResId id);
 public:
    ResManager();
    ~ResManager();

    void addProvider(ResProvider& provider);
    void addNewProvider(ResProvider* provider);
    void removeProvider(ResProvider& provider);

    void addPixmap(ResId id, Ptr<GfxPixmap> pix);
    void addSound(ResId id, SfxSampleData* snd);
    void addAnimation(ResId id, Ptr<GfxAnimation> anim);

    void clearPixmapCache();
    void clearSoundCache();
    void clearAnimationCache();

    Ptr<GfxPixmap>     getPixmap(ResId id);
    Ptr<GfxPixmap>     getPixmapT(ResId id);
    Ptr<GfxFont>       getFont(ResId id);
    Ptr<GfxAnimation>  getAnimation(ResId id);
    SfxSample          getSound(ResId id);
    Ptr<Stream>        getStream(ResId id);

    /** Get global resource manager instance. */
    static ResManager& getInstance()
        { return instance; }

    /// Object holder to hold resource providers.
    ObjectHolder h;
};

bool loadResource(ResManager&, string_t);
bool loadResourceConfig(ResManager&, Stream&);

/** Get sound for a resource Id from global resource manager.
    \return sample, might be null if no such sound.
    \see ResManager::getSound() */
inline SfxSample
requestSound(ResId id)
    { return ResManager::getInstance().getSound(id); }

/** Get pixmap for a resource Id from global resource manager.
    \return pixmap, might be null if no such pixmap.
    \see ResManager::getPixmap() */
inline Ptr<GfxPixmap>
requestPixmap(ResId id)
    { return ResManager::getInstance().getPixmap(id); }


#endif
