/**
  *  \file ui/widget-tab.cc
  *  \brief Register Cards
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "gfx/palette.h"
#include "util/misc.h"

/****************************** UICardGroup ******************************/

/** \class UICardGroup
    \brief Group containing widgets of which only one is visible at a time

    An UICardGroup can contain a number of child widgets with distinct
    widget Ids. Of these, only the focused widget is displayed; the
    others will be invisible. They will also not receive events.

    In order to change the displayed widget, change the focus.
    UICardGroup does not contain any provisions to do that. Another
    widget like UICardGroupTab must do that. */

/** Constructor. Build an empty card group. */
UICardGroup::UICardGroup()
{
    sig_focus_changed.add(this, &UICardGroup::onFocusChange);
}

UICardGroup::~UICardGroup()
{ }

void
UICardGroup::onFocusChange(int)
{
    drawWidget();
}

void
UICardGroup::drawEverything(GfxCanvas& can)
{
    drawContent(can);
    if (UIBaseWidget* w = getFocusedChild())
        w->drawEverything(can);
}

void
UICardGroup::drawContent(GfxCanvas& canvas)
{
    getSkin().drawBackground(canvas, getExtent());
    canvas.update(getExtent());
}

void
UICardGroup::drawBelow(UIBaseWidget& /*child*/, GfxCanvas& canvas)
{
    /* We always draw everything even if the child is invisible, in case the
       child is trying to render the background into backing store */
    drawContent(canvas);
}

GfxCanvas&
UICardGroup::getCanvasFor(UIBaseWidget& child, ObjectHolder& h)
{
    if (&child == getFocusedChild())
        return getParent()->getCanvasFor(*this, h);
    else
        return h.add(new GfxNullCanvas());
}

bool
UICardGroup::handleEvent(const UIEvent& e, bool second_pass)
{
    if (UIBaseWidget* w = getFocusedChild())
        return w->handleEvent(e, second_pass);
    else
        return false;
}

void
UICardGroup::onAddRemoveChild(UIBaseWidget& /*the_widget*/, bool /*added*/)
{
    drawWidget();
}

void
UICardGroup::onChildResize(UIBaseWidget& the_widget, GfxRect /*oldsize*/)
{
    // our children should not change their size. Try the best we can.
    if (&the_widget == getFocusedChild())
        drawWidget();
    
}

void
UICardGroup::add(UIBaseWidget& w)
{
    w.setExtent(getExtent());
    addChildWidget(w, getLastChild());
}

void
UICardGroup::onResize(GfxRect /*oldsize*/, bool by_layout)
{
    for (UIBaseWidget* child = getFirstChild(); child != 0; child = child->getNextSibling()) {
        child->setExtent(getExtent(), true);
    }
    if (!by_layout)
        drawWidget();
}

void
UICardGroup::getLayoutInfo(LayoutInfo& info)
{
    LayoutInfo ch;

    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(1, 1);
    for (UIBaseWidget* child = getFirstChild(); child != 0; child = child->getNextSibling()) {
        child->getLayoutInfo(ch);
        info.flags &= ch.flags;
        if (ch.min_size.x > info.min_size.x)
            info.min_size.x = ch.min_size.x;
        if (ch.min_size.y > info.min_size.y)
            info.min_size.y = ch.min_size.y;
        if (ch.pref_size.x > info.pref_size.x)
            info.pref_size.x = ch.pref_size.x;
        if (ch.pref_size.y > info.pref_size.y)
            info.pref_size.y = ch.pref_size.y;
    }
}

/***************************** UICardGroupTab ****************************/

/** \class UICardGroupTab
    \brief Tab bar for a card group

    This provides a bar of clickable register tabs to control the
    focused widget of a UICardGroup. Users can click at the tabs, hit
    the tab hot-key, or use Tab/Shift-Tab to navigate through all
    pages.

    You must build up the tabs using addTab(). */

//! Information about a tab. \internal
struct UICardGroupTab::TabInfo {
    int         id;             ///< Widget id of tab.
    string_t    name;           ///< Name displayed on top of screen.
    UIKey       key;            ///< Hot-key.

    TabInfo(int id, const string_t name, UIKey key);
    ~TabInfo();
};

UICardGroupTab::TabInfo::TabInfo(int id, const string_t name, UIKey key)
    : id(id), name(name), key(key)
{ }

UICardGroupTab::TabInfo::~TabInfo()
{ }

/** Constructor.
    \param group  the UICardGroupTab we want to control. */
UICardGroupTab::UICardGroupTab(UICardGroup& group)
    : group(group),
      group_conn(observe(group.sig_focus_changed))
{ }

UICardGroupTab::~UICardGroupTab()
{
    group_conn.disconnect();
}

void
UICardGroupTab::drawContent(GfxCanvas& can)
{
    int x = getExtent().x;
    int top = getExtent().y;
    int bot = getExtent().y + getExtent().h;

    // figure out current selected item
    int cur_id = getCurrentTab();
    UISkin& s = group.getSkin();   // used to draw interior of tabs which
                                   // is visually inside the group
    s.startDraw();

    for (std::size_t i = 0; i < tabs.size(); ++i) {
        // first, draw lines at bottom
        int this_width = fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 20;
        int lwidth = tabs[i]->id == cur_id ? 10 : this_width+16;
        can.drawHLine(x, bot-2, x + lwidth, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(x, bot-1, x + lwidth, standard_colors[COLOR_GRAY],  SOLID_LINE, GFX_ALPHA_OPAQUE);
        if (tabs[i]->id == cur_id)
            s.drawBackground(can, GfxRect(x + 10, bot - 2, this_width+6, 2));

        // left side of tab
        x += 10;
        can.drawVLine(x-1, top, bot-2, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawVLine(x, top+1, bot-1, standard_colors[COLOR_GRAY], SOLID_LINE, GFX_ALPHA_OPAQUE);

        // top
        can.drawHLine(x, top, x + this_width, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(x, top+1, x + this_width, standard_colors[COLOR_GRAY], SOLID_LINE, GFX_ALPHA_OPAQUE);

        // content and rhs
        int cx = 0, he = bot - top - 2, lastl = 1, bx = x + this_width;
        for (int l = 1; l < he; ++l) {
            int w = 6 * l / he;
            if (w != cx) {
                drawSolidBar    (can, GfxRect(bx + cx, top + lastl, 2,  l - lastl), standard_colors[COLOR_BLACK]);
                if (lastl == 1 && l != 1)  // HACK HACK HACK
                    ++lastl;
                s.drawBackground(can, GfxRect(x+1,     top + lastl, cx+this_width-1, l - lastl));
                lastl = l;
                cx = w;
            }
        }
        s.drawBackground(can, GfxRect(x+1,     top + lastl, cx+this_width-1, he - lastl));
        drawSolidBar    (can, GfxRect(bx + cx, top + lastl, 2,  he - lastl), standard_colors[COLOR_BLACK]);

        s.setColor(UIColor::tc_Static);
        s.setSolid(false);
        s.outText(can, FONT_TITLE, x + 10, top + 1, tabs[i]->name);

        x += this_width+6;
    }

    int right = getExtent().x + getExtent().w;
    if (x < right) {
        can.drawHLine(x, bot-2, right, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(x, bot-1, right, standard_colors[COLOR_GRAY],  SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    can.update(getExtent());
}

void
UICardGroupTab::getLayoutInfo(LayoutInfo& li)
{
    int min_x = 0;
    for (std::size_t i = 0; i < tabs.size(); ++i)
        min_x += fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 36;

    li.flags = li_FlexibleH;
    li.pref_size = li.min_size = GfxPoint(min_x, font_heights[FONT_TITLE] + 3);
}

bool
UICardGroupTab::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* FIXME: I'm not certain whether should be here.
             pro: allows all sorts of Id sequences
             con: UIFocusIterator exists and works fine */
        // tab: next tab
        if (e.key.code == SDLK_TAB) {
            int x = getCurrentTabIndex() + 1;
            if (x >= int(tabs.size()))
                x = 0;
            return setCurrentTab(tabs[x]->id);
        }
        // shift-tab: previous tab
        if (e.key.code == ss_Shift + SDLK_TAB) {
            int x = getCurrentTabIndex() - 1;
            if (x < 0)
                x = tabs.size() - 1;
            return setCurrentTab(tabs[x]->id);
        }
        // letters atop the bar
        for (std::size_t i = 0; i < tabs.size(); ++i) {
            if ((e.key.code & ~ss_Alt) == tabs[i]->key)
                return setCurrentTab(tabs[i]->id);
        }
    } else if (e.type == ev_Mouse && isInside(GfxPoint(e.mouse.x, e.mouse.y)) && e.mouse.b) {
        int x = getExtent().x;
        for (std::size_t i = 0; i < tabs.size(); ++i) {
            int tw = fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 20;
            x += 10;
            if (x > e.mouse.x)
                break;
            if (e.mouse.x < x + tw)
                return setCurrentTab(tabs[i]->id);
            x += tw+6;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Add a tab.
    \param id     Id of the widget inside the controlled group.
    \param name   Name to display on the tab bar.
    \param key    Hot-key. The UICardGroupTab will accept both the
                  hot-key, as well as Alt+hot-key to select this tab.
                  The hot-key for a tab will only be recognized when
                  the tab in question is not already selected, so the
                  tab selected with 's' can itself use the hot-key 's'. */
void
UICardGroupTab::addTab(int id, const string_t& name, UIKey key)
{
    tabs.push_back(new TabInfo(id, name, key));
    drawWidget();
}

/** Add a tab. \overload */
void
UICardGroupTab::addTab(int id, const UIKeyString& key)
{
    addTab(id, key.getString(), key.getKey());
}

/** Get currently-selected tab Id. This returns the Id of the widget
    currently shown, -1 if none. */
int
UICardGroupTab::getCurrentTab() const
{
    if (UIBaseWidget* p = group.getFocusedChild())
        return p->getId();
    else
        return -1;
}

/** Set currently-selected tab. This brings the widget with the given
    Id to the front.
    \return true iff there was a change, false if not. */
bool
UICardGroupTab::setCurrentTab(int id)
{
    setSelectedWidget();

    UIBaseWidget* w = group.getWidgetById(id);
    if (!w)
        return false;
    if (group.getFocusedChild() && w->hasState(st_Focused))
        return false;
    group.setFocusedWidget(*w);
    return true;
}

/** Get index of currently-selected tab.
    \return index (0..getNumTabs()-1), or -1 if none selected. */
int
UICardGroupTab::getCurrentTabIndex() const
{
    int i = getCurrentTab();
    for (std::size_t x = 0; x < tabs.size(); ++x)
        if (tabs[x]->id == i)
            return x;
    return -1;
}

/** Set current tab by index.
    \param n tab to select (0..getNumTabs()-1, 0 is left-most).
    \see getCurrentTabIndex
    \see setCurrentTab */
bool
UICardGroupTab::setCurrentTabIndex(int n)
{
    return setCurrentTab(tabs[n]->id);
}
